from datetime import datetime
from argparse import ArgumentParser

from rich.table import Table
from rich.prompt import Prompt
from rich.console import Console
from rich.traceback import install

from .api import Client
from .storage import Storage, DEFAULT_DATA, DEFAULT_CONFIG_PATH

install(show_locals=True)


class CLI:
    def __init__(self, client=None, storage=None) -> None:
        self._client = client
        self._storage = storage
        self.console = Console()

    @property
    def client(self) -> Client:
        if not self._client:
            raise ValueError("Client not set!")
        return self._client

    @client.setter
    def client(self, client: Client):
        self._client = client

    @property
    def storage(self) -> Storage:
        if not self._storage:
            raise ValueError("Storage not set!")
        return self._storage

    @storage.setter
    def storage(self, storage: Storage):
        self._storage = storage

    def aliases(self):
        table = self.alias_table(self.storage.aliases)
        self.console.print(table)

    def login(self):
        self.client.otp()
        otp = self.get_otp()
        self.client.full_login(otp)
        self.storage.save_credentials(
            self.client.token,  # type: ignore
            self.client.access_token,  # type: ignore
            self.client.username,
        )
        self.console.print(
            f"[bold green]Succesfully[/bold green] logged in as [bold red]{self.client.username}@duck.com[/bold red]"
        )
        self.console.print(f"[bold green]-> {self.client.real_email}[/bold green]")

    def generate_alias(self):
        alias = self.client.generate_alias()
        self.storage.add_alias(alias)
        self.console.print(f"[bold green]{alias}@duck.com[/bold green]")

    def remove_alias(self, key_like):
        alias = self.storage.remove_alias(key_like)
        table = self.alias_table([alias])
        table.title = "Removed alias"
        self.console.print(table)

    def get_email(self, key_like: str, recipient_email: str):
        alias = key_like if "@" in key_like else self.storage.alias_getter(key_like)[1]["address"] + "@duck.com"
        s = recipient_email.replace("@", "_at_") + "_" + alias
        self.console.print(s)

    def logout(self):
        self.storage.data = {**DEFAULT_DATA}
        self.storage.save()

    def alias_table(self, aliases: list[dict]):
        table = Table()
        table.add_column("Index", style="bold yellow")
        table.add_column("Alias")
        table.add_column("Creation Date")
        for index, alias in enumerate(aliases):
            table.add_row(
                str(index),
                f"{alias['address']}@duck.com",
                datetime.fromtimestamp(alias["created"]).strftime("%c"),
            )
        return table

    def get_otp(self):
        return Prompt.ask(
            "Enter your [bold red]OTP[/bold red] code or url [italic](it may take a while for the email to arrive)[/italic]"
        )


def main():
    cli = CLI()

    parser = ArgumentParser("ddge", description="DuckDuckGo Email CLI")
    parser.add_argument(
        "-c", "--config", help="Custom config path.", default=DEFAULT_CONFIG_PATH
    )
    subparsers = parser.add_subparsers(title="actions", required=False)

    login = subparsers.add_parser(
        "login",
        help="Allows you to login using the OTP code and saves the credentials.",
    )
    login.add_argument("username", help="The @duck.com username.")
    login.set_defaults(func=cli.login)

    logout = subparsers.add_parser(
        "logout", help="Deletes the locally saved credentials."
    )
    logout.set_defaults(func=cli.logout)

    generate = subparsers.add_parser(
        "generate", help="The default action. Generates an alias"
    )
    generate.set_defaults(func=cli.generate_alias)

    aliases = subparsers.add_parser("aliases", help="Lists the created aliases.")
    aliases.set_defaults(func=cli.aliases)

    remove = subparsers.add_parser(
        "remove", help="Delete the alias from the local storage."
    )
    remove.add_argument(
        "key_like", help="index, full alias or alias username(the part before @)"
    )
    remove.set_defaults(func=cli.remove_alias, pass_=["key_like"])

    get_email = subparsers.add_parser(
        "get_email", help="Format the recipient address in a way that lets you send the email through duck.com"
    )
    get_email.add_argument(
        "key_like", help="index, full alias or alias username(the part before @)"
    )
    get_email.add_argument(
        "recipient_email"
    )
    get_email.set_defaults(func=cli.get_email, pass_=["key_like", "recipient_email"])

    args = parser.parse_args()

    storage = Storage(args.config)
    username = getattr(args, "username", None)
    if username:
        username = username.split("@")[0]
    client = Client(username or storage.username, storage.token, storage.access_token)

    cli.client = client
    cli.storage = storage

    if not hasattr(args, "func"):
        cli.generate_alias()
    else:
        args.func(
            **{
                k: v
                for k, v in args.__dict__.items()
                if k in getattr(args, "pass_", [])
            }
        )



from requests import session

USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/110.0"
API_BASE = "https://quack.duckduckgo.com/api"
OTP = "/auth/loginlink"  # ?user=user ; GET ; always returns {}
LOGIN = "/auth/login"  # ?otp=abc&user=user ; POST ; returns {"status": "authenticated", "token": "[...]", "user": "user"}
DASHBOARD = "/email/dashboard"  # GET ; returns {"invites":[],"stats":{"addresses_generated":12},"user":{"access_token":"[...]","cohort":"[..]","email":"real_email","username":"user"}}
GEN_EMAIL = "/email/addresses"  # POST ; returns {"address":"1234abcd"}


class Client:
    def __init__(
        self, username: str, token: str | None = None, access_token: str | None = None
    ) -> None:
        self.username = username
        self.token = token
        self.access_token = access_token
        self.generated_addresses: int | None = None  # will be None if not logged in
        self.real_email: str | None = None  # will be None if not logged in
        self.logged_in = all((token, access_token))
        self.session = session()
        self.headers = {
            "Origin": "https://duckduckgo.com",
            "Referer": "https://duckduckgo.com",
            "User-Agent": USER_AGENT,
        }

    def otp(self, username: str | None = None) -> bool:
        """Will use `self.username` if `username` not given"""
        params = {"user": username or self.username}
        req = self.session.get(f"{API_BASE}{OTP}", params=params, headers=self.headers)
        req.raise_for_status()
        return True

    def login(self, otp: str, username: str | None = None) -> dict:
        """Will use `self.username` if `username` not given
        otp can be either the phrase or url
        returns {"status": "authenticated", "token": "[...]", "user": "user"}"""
        if "https://" in otp:
            otp = otp.split("otp=")[-1].split("&")[0]
        else:
            otp = otp.replace(" ", "-")
        params = {"user": username or self.username, "otp": otp}
        req = self.session.get(
            f"{API_BASE}{LOGIN}", params=params, headers=self.headers
        )
        req.raise_for_status()
        d = req.json()
        return d["token"]

    def dashboard(self) -> dict:
        """returns `{"invites":[],"stats":{"addresses_generated":12},"user":{"access_token":"[...]","cohort":"[..]","email":"real_email","username":"user"}}`"""
        headers = {**self.headers, "Authorization": f"Bearer {self.token}"}
        req = self.session.get(f"{API_BASE}{DASHBOARD}", headers=headers)
        req.raise_for_status()
        d = req.json()
        return d

    def full_login(self, otp: str, username: str | None = None):
        token = self.login(otp, username)
        self.token = token
        res = self.dashboard()
        self.access_token = res['user']['access_token']
        self.real_email = res['user']['email']
        self.generated_addresses = res['stats']['addresses_generated']
        self.logged_in = True
        return True

    def generate_alias(self) -> str:
        headers = {**self.headers, "Authorization": f"Bearer {self.access_token}"}
        req = self.session.post(f"{API_BASE}{GEN_EMAIL}", headers=headers)
        req.raise_for_status()
        d = req.json()
        return d["address"]



import setuptools


def long_description():
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path):
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]


setuptools.setup(
    name="ddge",
    version='0.2.137.4',
    description="DuckDuckGo Email Protection CLI",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author="Zoey !",
    maintainer_email="cb98uzhd@duck.com",
    license="GNU General Public License v3.0",
    url="https://github.com/lzgirlcat/ddge",
    python_requires=">=3.7",
    entry_points={"console_scripts": ["ddge = ddge.cli:main"]},
    install_requires=parse_requirements_file("requirements.txt"),
    include_package_data=True,
    keywords=["email", "duckdcukgo", "protection", "ddg"],
    project_urls={
        "Source (GitHub)": "https://github.com/lzgirlcat/ddge",
        "Issue Tracker": "https://github.com/lzgirlcat/ddge/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



from time import time
from sys import platform
from json import dump, load
from os import makedirs, path

DEFAULT_DATA = {"token": None, "access_token": None, "username": None, "aliases": []}


def get_adequate_config_path():
    if platform == "darwin":
        # i dont even fucking know or want to know
        return "~/Library/Preferences/ddge/data.json"
    elif "win" in platform:
        # same with this
        return "%USERPROFILE%\\AppData\\Local\\ddge\\data.json"
    else:
        return "~/.config/ddge.json"


DEFAULT_CONFIG_PATH = get_adequate_config_path()


class Storage:
    def __init__(self, file_path: str = DEFAULT_CONFIG_PATH) -> None:
        expanded = path.expanduser(file_path)
        self.file_path = expanded
        if path.exists(self.file_path):
            with open(self.file_path, "r") as f:
                self.data = load(f)
        else:
            self.data = {**DEFAULT_DATA}
            self.save()

    def save(self):
        dir = path.dirname(self.file_path)
        if dir:
            if not path.exists(dir):
                makedirs(dir)
        with open(self.file_path, "w") as f:
            dump(self.data, f)

    @property
    def username(self) -> str:
        return self.data["username"]

    @property
    def aliases(self) -> list[dict]:
        return self.data["aliases"]

    @property
    def token(self) -> str:
        return self.data["token"]

    @property
    def access_token(self) -> str:
        return self.data["access_token"]

    def alias_getter(self, key_like: str):
        for index, alias in enumerate(self.aliases):
            if key_like == str(index):
                return index, alias
            elif key_like == alias["address"]:
                return index, alias
            elif key_like.split("@")[0] == alias["address"]:
                return index, alias
        return None, None

    def remove_alias(self, key_like: str):
        """`key_like` accepts the index, full alias and alias username(the part before @)"""
        index, ret = self.alias_getter(key_like)

        if index is not None:
            self.data["aliases"].pop(index)
            self.save()
            return ret
        raise KeyError(f"Didn't find any aliases matching '{key_like}'")

    def add_alias(self, address):
        alias = {"address": address, "created": int(time())}
        self.data["aliases"].append(alias)
        self.save()
        return alias

    def save_credentials(self, token: str, access_token: str, username: str):
        self.data["token"] = token
        self.data["access_token"] = access_token
        self.data["username"] = username
        self.save()
